"""Tests related to petab.conditions"""
import os
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import petab
from petab import conditions
from petab.C import *


def test_get_parametric_overrides():
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )

    assert conditions.get_parametric_overrides(condition_df) == []

    condition_df.fixedParameter1 = condition_df.fixedParameter1.values.astype(
        int
    )

    assert conditions.get_parametric_overrides(condition_df) == []

    condition_df["fixedParameter1"] = condition_df["fixedParameter1"].astype(
        "object"
    )
    condition_df.loc[0, "fixedParameter1"] = "parameterId"

    assert conditions.get_parametric_overrides(condition_df) == ["parameterId"]


def test_get_condition_df():
    """Test conditions.get_condition_df."""
    # condition df missing ids
    condition_df = pd.DataFrame(
        data={
            CONDITION_NAME: ["Condition 1", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        condition_df.to_csv(fh, sep="\t", index=False)

    with pytest.raises(KeyError):
        petab.get_condition_df(file_name)

    os.remove(file_name)

    # with ids
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        condition_df.to_csv(fh, sep="\t", index=False)

    df = petab.get_condition_df(file_name).replace(np.nan, "")
    assert (df == condition_df.set_index(CONDITION_ID)).all().all()

    os.remove(file_name)

    # test other arguments
    assert (petab.get_condition_df(condition_df) == condition_df).all().all()
    assert petab.get_condition_df(None) is None


def test_write_condition_df():
    """Test conditions.write_condition_df."""
    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["Condition 1", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    ).set_index(CONDITION_ID)

    with tempfile.TemporaryDirectory() as temp_dir:
        file_name = Path(temp_dir) / "conditions.tsv"
        petab.write_condition_df(condition_df, file_name)
        re_df = petab.get_condition_df(file_name)
        assert (condition_df == re_df).all().all()


def test_create_condition_df():
    """Test conditions.create_condition_df."""
    parameter_ids = ["par1", "par2", "par3"]
    condition_ids = ["condition1", "condition2"]

    df = petab.create_condition_df(parameter_ids, condition_ids)

    expected = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            "par1": [np.nan, np.nan],
            "par2": [np.nan, np.nan],
            "par3": [np.nan, np.nan],
        }
    ).set_index(CONDITION_ID)

    assert ((df == expected) | df.isna() == expected.isna()).all().all()



"""Validation of PEtab visualization files"""
import logging

import pandas as pd

from .. import C, Problem
from ..C import VISUALIZATION_DF_REQUIRED_COLS

logger = logging.getLogger(__name__)


def validate_visualization_df(problem: Problem) -> bool:
    """Validate visualization table

    Arguments:
        problem: The PEtab problem containing a visualization table

    Returns:
        ``True`` if errors occurred, ``False`` otherwise
    """
    vis_df = problem.visualization_df
    if vis_df is None or vis_df.empty:
        return False

    errors = False

    if missing_req_cols := (
        set(VISUALIZATION_DF_REQUIRED_COLS) - set(vis_df.columns)
    ):
        logger.error(
            f"Missing required columns {missing_req_cols} "
            "in visualization table."
        )
        errors = True

    # Set all unspecified optional values to their defaults to simplify
    # validation
    vis_df = vis_df.copy()
    _apply_defaults(vis_df)

    if unknown_types := (
        set(vis_df[C.PLOT_TYPE_SIMULATION].unique())
        - set(C.PLOT_TYPES_SIMULATION)
    ):
        logger.error(
            f"Unknown {C.PLOT_TYPE_SIMULATION}: {unknown_types}. "
            f"Must be one of {C.PLOT_TYPES_SIMULATION}"
        )
        errors = True

    if unknown_types := (
        set(vis_df[C.PLOT_TYPE_DATA].unique()) - set(C.PLOT_TYPES_DATA)
    ):
        logger.error(
            f"Unknown {C.PLOT_TYPE_DATA}: {unknown_types}. "
            f"Must be one of {C.PLOT_TYPES_DATA}"
        )
        errors = True

    if unknown_scale := (set(vis_df[C.X_SCALE].unique()) - set(C.X_SCALES)):
        logger.error(
            f"Unknown {C.X_SCALE}: {unknown_scale}. "
            f"Must be one of {C.X_SCALES}"
        )
        errors = True

    if any(
        (vis_df[C.X_SCALE] == "order")
        & (vis_df[C.PLOT_TYPE_SIMULATION] != C.LINE_PLOT)
    ):
        logger.error(
            f"{C.X_SCALE}=order is only allowed with "
            f"{C.PLOT_TYPE_SIMULATION}={C.LINE_PLOT}."
        )
        errors = True

    if unknown_scale := (set(vis_df[C.Y_SCALE].unique()) - set(C.Y_SCALES)):
        logger.error(
            f"Unknown {C.Y_SCALE}: {unknown_scale}. "
            f"Must be one of {C.Y_SCALES}"
        )
        errors = True

    if problem.condition_df is not None:
        # check for ambiguous values
        reserved_names = {C.TIME, "condition"}
        for reserved_name in reserved_names:
            if (
                reserved_name in problem.condition_df
                and reserved_name in vis_df[C.X_VALUES]
            ):
                logger.error(
                    f"Ambiguous value for `{C.X_VALUES}`: "
                    f"`{reserved_name}` has a special meaning as "
                    f"`{C.X_VALUES}`, but there exists also a model "
                    "entity with that name."
                )
                errors = True

        # check xValues exist in condition table
        for xvalue in set(vis_df[C.X_VALUES].unique()) - reserved_names:
            if xvalue not in problem.condition_df:
                logger.error(
                    f"{C.X_VALUES} was set to `{xvalue}`, but no "
                    "such column exists in the conditions table."
                )
                errors = True

        if problem.observable_df is not None:
            # yValues must be an observable
            for yvalue in vis_df[C.Y_VALUES].unique():
                if pd.isna(yvalue):
                    # if there is only one observable, we default to that
                    if len(problem.observable_df.index.unique()) == 1:
                        continue

                    logger.error(
                        f"{C.Y_VALUES} must be specified if there is more "
                        "than one observable."
                    )
                    errors = True

                if yvalue not in problem.observable_df.index:
                    logger.error(
                        f"{C.Y_VALUES} was set to `{yvalue}`, but no such "
                        "observable exists in the observables table."
                    )
                    errors = True

    return errors


def _apply_defaults(vis_df: pd.DataFrame):
    """
    Set default values.

    Adds default values to the given visualization table where no value was
    specified.
    """

    def set_default(column: str, value):
        if column not in vis_df:
            vis_df[column] = value
        elif value is not None:
            if isinstance(value, str):
                vis_df[column] = vis_df[column].astype("object")
            vis_df.fillna({column: value}, inplace=True)

    set_default(C.PLOT_NAME, "")
    set_default(C.PLOT_TYPE_SIMULATION, C.LINE_PLOT)
    set_default(C.PLOT_TYPE_DATA, C.MEAN_AND_SD)
    set_default(C.DATASET_ID, None)
    set_default(C.X_VALUES, C.TIME)
    set_default(C.X_OFFSET, 0)
    set_default(C.X_LABEL, vis_df[C.X_VALUES])
    set_default(C.X_SCALE, C.LIN)
    set_default(C.Y_VALUES, None)
    set_default(C.Y_OFFSET, 0)
    set_default(C.Y_LABEL, vis_df[C.Y_VALUES])
    set_default(C.Y_SCALE, C.LIN)
    set_default(C.LEGEND_ENTRY, vis_df[C.DATASET_ID])



"""Tests for petab.simplify.*"""
from math import nan

import pandas as pd
import pytest
import simplesbml
from pandas.testing import *

from petab import Problem
from petab.C import *  # noqa: F403
from petab.models.sbml_model import SbmlModel
from petab.simplify import *


@pytest.fixture
def problem() -> Problem:
    ss_model = simplesbml.SbmlModel()
    ss_model.addParameter("some_parameter", val=1.0)
    ss_model.addParameter("same_value_for_all_conditions", val=1.0)

    observable_df = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs_used", "obs_unused", "obs_used_2"],
            OBSERVABLE_FORMULA: [1.0, 2.0, 3.0],
            NOISE_FORMULA: [1.0, 2.0, 3.0],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    conditions_df = pd.DataFrame(
        {
            CONDITION_ID: [
                "condition_used_1",
                "condition_unused",
                "condition_used_2",
            ],
            "some_parameter": [1.0, 2.0, 3.0],
            "same_value_for_all_conditions": [4.0] * 3,
        }
    )
    conditions_df.set_index(CONDITION_ID, inplace=True)

    measurement_df = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs_used", "obs_used_2", "obs_used"],
            MEASUREMENT: [1.0, 1.5, 2.0],
            SIMULATION_CONDITION_ID: [
                "condition_used_1",
                "condition_used_1",
                "condition_used_2",
            ],
            TIME: [1.0] * 3,
        }
    )
    yield Problem(
        model=SbmlModel(sbml_model=ss_model.getModel()),
        condition_df=conditions_df,
        observable_df=observable_df,
        measurement_df=measurement_df,
    )


def test_remove_nan_measurements(problem):
    expected = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs_used"] * 2,
            MEASUREMENT: [1.0, 2.0],
            SIMULATION_CONDITION_ID: ["condition_used_1", "condition_used_2"],
            TIME: [1.0] * 2,
        }
    )

    problem.measurement_df = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs_used", "obs_with_nan", "obs_used"],
            MEASUREMENT: [1.0, nan, 2.0],
            SIMULATION_CONDITION_ID: [
                "condition_used_1",
                "condition_used_1",
                "condition_used_2",
            ],
            TIME: [1.0] * 3,
        }
    )
    assert not problem.measurement_df.equals(expected)

    remove_nan_measurements(problem)

    assert_frame_equal(problem.measurement_df, expected)


def test_remove_unused_observables(problem):
    expected = pd.DataFrame(
        {
            OBSERVABLE_ID: ["obs_used", "obs_used_2"],
            OBSERVABLE_FORMULA: [1.0, 3.0],
            NOISE_FORMULA: [1.0, 3.0],
        }
    )
    expected.set_index(OBSERVABLE_ID, inplace=True)
    assert not problem.observable_df.equals(expected)

    remove_unused_observables(problem)

    assert_frame_equal(problem.observable_df, expected)


def test_remove_unused_conditions(problem):
    expected = pd.DataFrame(
        {
            CONDITION_ID: ["condition_used_1", "condition_used_2"],
            "some_parameter": [1.0, 3.0],
            "same_value_for_all_conditions": [4.0] * 2,
        }
    )
    expected.set_index(CONDITION_ID, inplace=True)
    assert not problem.condition_df.equals(expected)

    remove_unused_conditions(problem)

    assert_frame_equal(problem.condition_df, expected)


def test_condition_parameters_to_parameter_table(problem):
    expected_conditions = pd.DataFrame(
        {
            CONDITION_ID: [
                "condition_used_1",
                "condition_unused",
                "condition_used_2",
            ],
            "some_parameter": [1.0, 2.0, 3.0],
        }
    )
    expected_conditions.set_index(CONDITION_ID, inplace=True)
    assert not problem.condition_df.equals(expected_conditions)

    expected_parameters = pd.DataFrame(
        {
            PARAMETER_ID: ["same_value_for_all_conditions"],
            PARAMETER_SCALE: [LIN],
            LOWER_BOUND: [nan],
            UPPER_BOUND: [nan],
            NOMINAL_VALUE: [4.0],
            ESTIMATE: [0],
        }
    )
    expected_parameters.set_index(PARAMETER_ID, inplace=True)
    assert problem.parameter_df is None

    condition_parameters_to_parameter_table(problem)

    assert_frame_equal(problem.condition_df, expected_conditions)
    assert_frame_equal(problem.parameter_df, expected_parameters)


def test_simplify_problem(problem):
    # simplify_problem checks whether the result is valid
    simplify_problem(problem)



#!/usr/bin/env python3

"""Command line tool to check for correct format"""

import argparse
import logging
import sys

from colorama import Fore
from colorama import init as init_colorama

import petab

logger = logging.getLogger(__name__)


class LintFormatter(logging.Formatter):
    """Custom log formatter"""

    formats = {
        logging.DEBUG: Fore.CYAN + "%(message)s",
        logging.INFO: Fore.GREEN + "%(message)s",
        logging.WARN: Fore.YELLOW + "%(message)s",
        logging.ERROR: Fore.RED + "%(message)s",
    }

    def format(self, record):
        # pylint: disable=protected-access
        format_orig = self._style._fmt
        self._style._fmt = LintFormatter.formats.get(record.levelno, self._fmt)
        result = logging.Formatter.format(self, record)
        self._style._fmt = format_orig
        return result


def parse_cli_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description="Check if a set of files adheres to the PEtab format."
    )

    # General options:
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="More verbose output",
    )

    # Call with set of files
    parser.add_argument(
        "-s", "--sbml", dest="sbml_file_name", help="SBML model filename"
    )
    parser.add_argument(
        "-o",
        "--observables",
        dest="observable_file_name",
        help="Observable table",
    )
    parser.add_argument(
        "-m",
        "--measurements",
        dest="measurement_file_name",
        help="Measurement table",
    )
    parser.add_argument(
        "-c",
        "--conditions",
        dest="condition_file_name",
        help="Conditions table",
    )
    parser.add_argument(
        "-p",
        "--parameters",
        dest="parameter_file_name",
        help="Parameter table",
    )
    parser.add_argument(
        "--vis",
        "--visualizations",
        dest="visualization_file_name",
        help="Visualization table",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-y",
        "--yaml",
        dest="yaml_file_name",
        help="PEtab YAML problem filename",
    )

    args = parser.parse_args()

    if args.yaml_file_name and any(
        (
            args.sbml_file_name,
            args.condition_file_name,
            args.measurement_file_name,
            args.parameter_file_name,
        )
    ):
        parser.error(
            "When providing a yaml file, no other files may " "be specified."
        )

    return args


def main():
    """Run PEtab validator"""
    args = parse_cli_args()
    init_colorama(autoreset=True)

    ch = logging.StreamHandler()
    if args.verbose:
        ch.setLevel(logging.DEBUG)
    else:
        ch.setLevel(logging.WARN)
    ch.setFormatter(LintFormatter())
    logging.basicConfig(level=logging.DEBUG, handlers=[ch])

    if args.yaml_file_name:
        from jsonschema.exceptions import ValidationError

        from petab.yaml import validate

        try:
            validate(args.yaml_file_name)
        except ValidationError as e:
            logger.error(
                "Provided YAML file does not adhere to PEtab " f"schema: {e}"
            )
            sys.exit(1)

        if petab.is_composite_problem(args.yaml_file_name):
            # TODO: further checking:
            #  https://github.com/ICB-DCM/PEtab/issues/191
            #  problem = petab.CompositeProblem.from_yaml(args.yaml_file_name)
            return

        problem = petab.Problem.from_yaml(args.yaml_file_name)

    else:
        logger.debug("Looking for...")
        if args.sbml_file_name:
            logger.debug(f"\tSBML model: {args.sbml_file_name}")
        if args.condition_file_name:
            logger.debug(f"\tCondition table: {args.condition_file_name}")
        if args.observable_file_name:
            logger.debug(f"\tObservable table: {args.observable_file_name}")
        if args.measurement_file_name:
            logger.debug(f"\tMeasurement table: {args.measurement_file_name}")
        if args.parameter_file_name:
            logger.debug(f"\tParameter table: {args.parameter_file_name}")
        if args.visualization_file_name:
            logger.debug(
                "\tVisualization table: " f"{args.visualization_file_name}"
            )

        try:
            problem = petab.Problem.from_files(
                sbml_file=args.sbml_file_name,
                condition_file=args.condition_file_name,
                measurement_file=args.measurement_file_name,
                parameter_file=args.parameter_file_name,
                observable_files=args.observable_file_name,
                visualization_files=args.visualization_file_name,
            )
        except FileNotFoundError as e:
            logger.error(e)
            sys.exit(1)

    ret = petab.lint.lint_problem(problem)
    sys.exit(ret)


if __name__ == "__main__":
    main()
